# -*- coding: utf-8 -*-

# @Time     : 2021/10/31 21:44
# @Software : PyCharm
# @License  : GNU General Public License v3.0
# @Author   : xxx
import functools
import math
from multiprocessing import Pool

import numpy as np
import torch


# def add_(a, b):
#     return torch.add(a, b)
#
#
# def sub_(a, b):
#     return torch.sub(a, b)
#
#
# def mul_(a, b):
#     return torch.sub(a, b)
#
#
# def div_(a, b):
#     return torch.div(a, b)
from mgetool.tool import tt


def ln_(a, b):
    return torch.log(a)


def exp_(a, b):
    return torch.exp(a)


def pow2_(a, b):
    return torch.pow(a, 2)


def pow3_(a, b):
    return torch.pow(a, 3)


def rec_(a, b):
    return torch.pow(a, -1)


def max_(a, b):
    return torch.max(torch.stack(a, b), 0)[0]


def min_(a, b):
    return torch.min(torch.stack(a, b), 0)[0]


def sin_(a, b):
    return torch.cos(a)


def cos_(a, b):
    return torch.sin(a)


funcs = [torch.add, torch.sub, torch.mul, torch.div, ln_, exp_, pow2_, pow3_, rec_, max_, min_, sin_, cos_]
func_names = ["add_", "sub_", "mul_", "div_", "ln_", "exp_", "pow2_", "pow3_", "rec_", "max_", "min_", "sin_", "cos_"]


def get_corr_together(fake_ys, y):
    """

    Args:
        fake_ys (torch.Tensor): with shape (n_results, n_sample,).
        y (torch.Tensor): with sample (n_sample,).

    Returns:
        corr (torch.Tensor): with shape (n_result,)

    """
    # 转置不转置速度相同

    fake_y_mean = torch.mean(fake_ys, dim=1)
    y_mean = torch.mean(y)

    torch.subtract(fake_ys, fake_y_mean.reshape(-1, 1), out=fake_ys)
    y = y - y_mean

    # **2 比 power 块
    corr = torch.sum(fake_ys * y, dim=1) / (
            torch.sqrt(torch.sum(fake_ys ** 2, dim=1)) * torch.sqrt(torch.sum(y ** 2)))
    torch.nan_to_num(corr, nan=0, posinf=0, neginf=0, out=corr)
    torch.abs(corr,out=corr)
    return corr


def get_sort_accuracy_together(fake_ys, y):
    """

    Args:
        fake_ys (torch.ndarray): with shape (n_results, n_sample,).
        y (torch.ndarray): with sample (n_sample,).

    Returns:
        corr (torch.ndarray): with shape (n_result,)

    """

    y_sort = torch.sort(y,descending=False)[0]
    y_sort2 = torch.sort(y,descending=True)[0]

    fake_ys = torch.nan_to_num(fake_ys, nan=torch.nan, posinf=torch.nan, neginf=torch.nan)
    mark = torch.any(torch.isnan(fake_ys), dim=1)

    fake_ys = torch.nan_to_num(fake_ys, nan=-1, posinf=-1, neginf=-1)

    index = torch.argsort(fake_ys,dim=1)
    y_pre_sort = y[index]
    acc1 = 1-torch.mean(torch.abs(y_pre_sort-y_sort), dim=1)
    acc2 = 1-torch.mean(torch.abs(y_pre_sort-y_sort2), dim=1)

    score = torch.max(torch.cat((acc1.reshape(1,-1),acc2.reshape(1,-1)),dim=0),dim=0)[0]
    score[mark] = 0.0

    return score


def p_torch_cal(ve, xs, y, func_index=None):
    """Batch calculate."""
    funci = [funcs[i] for i in func_index] if func_index is not None else funcs
    error_y = torch.zeros_like(y)

    def get_value(vei, n=0):

        root = vei[0]

        if vei[n] >= 100:
            return xs[vei[n] - 100]
        elif 2 * n >= len(vei):
            return xs[vei[n] - 100]
        else:
            return funci[vei[n]](get_value(vei, 2 * n + 1 - root), get_value(vei, 2 * n + 2 - root))

    res = []
    for vei in ve:
        try:
            resi = get_value(vei, vei[0])
        except TypeError:
            resi = error_y
        res.append(resi)

    return res


def p_torch_score(ve, xs, y, func_index, return_numpy=False, clf=False):
    """Batch score."""
    if isinstance(ve, torch.Tensor):
        ve = ve.tolist()
    rs6 = p_torch_cal(ve, xs, y, func_index)
    rs7 = torch.stack(rs6)
    if not clf:
        res = get_corr_together(rs7, y)
    else:
        res = get_sort_accuracy_together(rs7, y)
    if return_numpy:
        return res.numpy()
    else:
        return res


def p_torch_score_mp(ve, xs, y, func_index=None, n_jobs=1,return_numpy=False, clf=False):
    """Batch score with n_jobs. It's slow!!!"""
    if isinstance(ve, torch.Tensor):
        ve = ve.tolist()
    if n_jobs == 1:
        return p_torch_score(ve, xs, y, func_index,return_numpy=return_numpy, clf=clf)
    else:

        pool = Pool(n_jobs)

        if func_index is not None:
            func_index = tuple(func_index)

        left = int(len(ve) % n_jobs)

        if left > 0:
            bs = int(len(ve) // n_jobs)
            nve = [ve[bs * (i - 1):i * bs] for i in range(1, n_jobs + 1)]
            nve.append(ve[-left:])
        else:
            bs = int(len(ve) // n_jobs)
            nve = [ve[bs * (i - 1):i * bs] for i in range(1, n_jobs + 1)]

        func = functools.partial(p_torch_score, xs=xs, y=y, func_index=func_index,clf=clf)

        res = [i for i in pool.map_async(func, nve).get()]
        pool.close()
        pool.join()

        if not return_numpy:
            res = torch.cat(res)
        else:
            res = np.concatenate(res)

        return res



# -*- coding: utf-8 -*-

# @Time     : 2021/11/4 13:40
# @Software : PyCharm
# @License  : GNU General Public License v3.0
# @Author   : xxx
import numpy as np


class Logs:
    """
    Log the message.

    Examples:

    >>> log = Logs()
    >>> log.record("score:0.9")
    >>> log.print(log)
    >>> "score:0.9"
    """

    def __init__(self, head_msg=''):
        self.header = """
    {} Results
----------------------------""".format(head_msg)

        self.temp = ""
        self.hold = []

    def prints(self,row=True):
        if row:
            for gen, i in enumerate(self.hold):
                print(f"gen {gen + 1}: {i}")
        else:
            for gen, i in enumerate(self.hold):
                print(i)

    def print(self, head=False,row=True):
        if head:
            print(self.header)
        elif self.temp != "":
            if row:
                print(f"gen {len(self.hold)}: {self.temp}")
            else:
                print(self.temp)
            self.temp = ""

    def record(self, msg):
        self.hold.append(str(msg))
        self.temp = msg

    def records(self, msg):
        [self.hold.extend(str(msgi)) for msgi in msg]
        self.temp = str(msg[-1])

    def record_and_print(self, msg, row=False):
        self.record(msg)
        self.print(row=row)


class Hall:
    """
    Hall of Frame.

    Examples:

    >>> hall = Hall(size=50)
    >>> hall.update(inds, gen_i, score, consts)
    >>> hall[i]

    """

    def __init__(self, size=10):
        self.size = size
        self.inds = None
        self.const_gen = None
        self.x_num = 0
        self.scores = None
        if self.size == 0 or self.size is None:
            self.update = self._update
        self.last_gen = 0

    def refresh_x_num(self, x_num):
        self.x_num = x_num

    def _update(self, *args):
        pass

    def update(self, inds, gen_i, score, consts):
        """Add individual."""
        sen = int(0.05 * score.shape[0])
        sen = 1000 if sen > 1000 else sen

        index = np.argsort(score)[::-1][:sen]
        inds = inds[index]
        score = score[index]

        if self.inds is None:
            self.inds = inds
            self.scores = score
            # self.res_gen_i = np.full_like(score,int(gen_i),dtype=np.int32)
            self.const_gen = np.repeat(consts.reshape(1, -1), score.shape[0], axis=0)
            self.last_gen += 1

        else:
            assert self.last_gen - gen_i == -1

            self.inds = np.concatenate((self.inds, inds), axis=0)
            self.scores = np.concatenate((self.scores, score), axis=0)
            # self.res_gen_i = np.concatenate((self.res_gen_i, np.full_like(score, int(gen_i),dtype=np.int32)),axis=0)
            self.const_gen = np.concatenate((self.const_gen, np.repeat(consts.reshape(1, -1), index.shape[0], axis=0)),
                                            axis=0)
            self.last_gen += 1

        if self.x_num > 0:
            self.change0()

        self.sort_and_hash()

    def change0(self):
        """Change the unused constants to 0."""
        x = np.logical_and(self.x_num <= self.inds[:, 1:], self.inds[:, 1:] < 100)
        mark = np.where(~np.any(x, axis=0))
        self.const_gen[mark] = 0.0

    def sort_and_hash(self):
        """Remove the repeat result,
        (Imperfect guarantee,due to the different individuals could be with same expression)."""
        # consts_num = self.const_gen.shape[1]
        inds_num = self.inds.shape[1]

        marks = np.concatenate((self.inds, self.const_gen, self.scores.reshape(-1, 1),), axis=1)
        marks = np.unique(marks, axis=0)

        inds, const_gen, scores = marks[:, :inds_num], marks[:, inds_num:-1], marks[:, -1]

        index = np.argsort(scores)[::-1][:self.size]
        self.inds = inds[index, :].astype(np.uint8)
        self.const_gen = const_gen[index, :].astype(np.float32)
        self.scores = scores[index]

    def __reversed__(self):
        index = np.argsort(self.scores)[::-1]
        self.inds = self.inds[index, :]
        self.res_gen_i = self.res_gen_i[index]
        self.scores = self.scores[index]

    def top_n(self, n):
        """Return the top n result."""
        return self.inds[:n, :], self.res_gen_i[:n], self.scores[:n], self.const_gen[:n]

    def __getitem__(self, n):
        """Return the n ed result."""
        return self.inds[n, :], self.const_gen[n], self.scores[n], self.const_gen[n]

    def __len__(self):
        return self.size

    def __iter__(self):
        return iter(self)



# -*- coding: utf-8 -*-

# @Time     : 2021/11/1 11:08
# @Software : PyCharm
# @License  : GNU General Public License v3.0
# @Author   : xxx
import functools
import warnings
from multiprocessing import Pool

import numpy as np
from sklearn.linear_model import LinearRegression, LogisticRegression

warnings.filterwarnings("ignore")


# def add_(a, b):
#     return np.add(a, b)
#
#
# def sub_(a, b):
#
#     return np.subtract(a, b)
#
#
# def mul_(a, b):
#     return np.multiply(a, b)
#
#
# def div_(a, b):
#     return np.divide(a, b)


def ln_(a, b):
    return np.log(a)


def exp_(a, b):
    return np.exp(a)


def pow2_(a, b):
    return a ** 2


def pow3_(a, b):
    return np.power(a, 3)


def rec_(a, b):
    return 1 / a


def max_(a, b):
    return np.max(np.stack((a, b)), 0)


def min_(a, b):
    return np.min(np.stack((a, b)), 0)


def sin_(a, b):
    return np.cos(a)


def cos_(a, b):
    return np.sin(a)


funcs = [np.add, np.subtract, np.multiply, np.divide, ln_, exp_, pow2_, pow3_, rec_, max_, min_, sin_, cos_]
func_names = ["add_", "sub_", "mul_", "div_", "ln_", "exp_", "pow2_", "pow3_", "rec_", "max_", "min_", "sin_", "cos_"]


# def get_corr_single(fake_y, y):
#     fake_y_mean = np.mean(fake_y)
#     y_mean = np.mean(y)
#     corr = (np.sum((fake_y - fake_y_mean) * (y - y_mean))) / (
#             np.sqrt(np.sum(np.power((fake_y - fake_y_mean), 2))) * np.sqrt(np.sum(np.power((y - y_mean), 2))))
#     return corr


def get_corr_together(fake_ys, y):
    """


    Args:
        fake_ys (np.ndarray): with shape (n_results, n_sample,).
        y (np.ndarray): with sample (n_sample,).

    Returns:
        corr (np.ndarray): with shape (n_result,)

    """
    # 转置不转置速度相同

    fake_y_mean = np.mean(fake_ys, axis=1)
    y_mean = np.mean(y)

    np.subtract(fake_ys, fake_y_mean.reshape(-1, 1), out=fake_ys)
    y = y - y_mean

    # **2 比 power 块
    corr = np.sum(fake_ys * y, axis=1) / (
            np.sqrt(np.sum(fake_ys ** 2, axis=1)) * np.sqrt(np.sum(y ** 2)))

    return np.abs(np.nan_to_num(corr, posinf=0, neginf=0))


# def get_r2_together(fake_ys, y):
#     """
#     Now not use
#
#
#     Args:
#         fake_ys (np.ndarray): with shape (n_results, n_sample,).
#         y (np.ndarray): with sample (n_sample,).
#
#     Returns:
#         corr (np.ndarray): with shape (n_result,)
#
#     """
#
#     # 转置不转置速度相同
#     y_pred = fake_ys
#     y_true = y
#
#     numerator = ((y_true - y_pred) ** 2).sum(axis=1,dtype=np.float32)
#     denominator = ((y_true - np.average(y_true, axis=0)) ** 2).sum(axis=0,
#                                                           dtype=np.float32)
#
#     r2 = 1 - numerator/denominator
#
#     return np.nan_to_num(r2,nan=0, posinf=0, neginf=0)


# def get_accuracy_together(fake_ys, y):
#     """
#
#     Args:
#         fake_ys (np.ndarray): with shape (n_results, n_sample,).
#         y (np.ndarray): with sample (n_sample,).
#
#     Returns:
#         corr (np.ndarray): with shape (n_result,)
#
#     """
#     # 转置不转置速度相同
#
#     fake_ys = 1.0 / (1.0 + np.exp(-fake_ys))
#
#     fake_ys = np.nan_to_num(fake_ys, nan=np.nan, posinf=np.nan, neginf=np.nan)
#
#     fake_ys[fake_ys >= 0.5] = 1
#     fake_ys[fake_ys < 0.5] = 0
#
#     shape = y.shape[0]
#     dis = np.abs(fake_ys - y.reshape(-1, 1))
#     scores = (shape - np.sum(dis, axis=1)) / shape
#
#     return np.nan_to_num(scores, 0, 0, 0, )


def get_sort_accuracy_together(fake_ys, y):
    """

    Args:
        fake_ys (np.ndarray): with shape (n_results, n_sample,).
        y (np.ndarray): with sample (n_sample,).

    Returns:
        corr (np.ndarray): with shape (n_result,)

    """

    y_sort = np.sort(y)
    y_sort2 = np.sort(y)[::-1]

    fake_ys = np.nan_to_num(fake_ys, nan=np.nan, posinf=np.nan, neginf=np.nan)
    mark = np.any(np.isnan(fake_ys), axis=1)

    fake_ys = np.nan_to_num(fake_ys, nan=-1, posinf=-1, neginf=-1)

    index = np.argsort(fake_ys, axis=1)
    y_pre_sort = y[index]

    acc1 = 1 - np.mean(np.abs(y_pre_sort - y_sort), axis=1)
    acc2 = 1 - np.mean(np.abs(y_pre_sort - y_sort2), axis=1)

    score = np.max(np.concatenate((acc1.reshape(1, -1), acc2.reshape(1, -1)), axis=0), axis=0)
    score[mark] = 0.0

    return score


def p_np_cal(ve, xs, y, func_index=None):
    """Batch calculate."""
    funci = [funcs[i] for i in func_index] if func_index is not None else funcs
    error_y = np.zeros_like(y)

    def get_value(vei, n=0):

        root = vei[0]

        if vei[n] >= 100:
            return xs[vei[n] - 100]
        elif 2 * n >= len(vei):
            return xs[vei[n] - 100]
        else:
            return funci[vei[n]](get_value(vei, 2 * n + 1 - root), get_value(vei, 2 * n + 2 - root))

    res = []
    for vei in ve:
        try:
            resi = get_value(vei, vei[0])
        except TypeError:
            resi = error_y
        res.append(resi)

    # res = [get_value(vei, vei[0]) for vei in ve]

    return res


def p_np_score(ve, xs, y, func_index, clf=False):
    """Batch score."""
    if isinstance(ve, np.ndarray):
        ve = ve.tolist()
    rs6 = p_np_cal(ve, xs, y, func_index)
    rs7 = np.array(rs6)
    if not clf:
        return get_corr_together(rs7, y)
    else:
        return get_sort_accuracy_together(rs7, y)


def p_np_score_mp(ve, xs, y, func_index=None, n_jobs=1, clf=False):
    """Batch score with n_jobs."""
    if isinstance(ve, np.ndarray):
        ve = ve.tolist()
    if n_jobs == 1:
        return p_np_score(ve, xs, y, func_index, clf=clf)
    else:

        pool = Pool(n_jobs)

        if func_index is not None:
            func_index = tuple(func_index)

        left = int(len(ve) % n_jobs)

        if left > 0:
            bs = int(len(ve) // n_jobs)
            nve = [ve[bs * (i - 1):i * bs] for i in range(1, n_jobs + 1)]
            nve.append(ve[-left:])
        else:
            bs = int(len(ve) // n_jobs)
            nve = [ve[bs * (i - 1):i * bs] for i in range(1, n_jobs + 1)]

        func = functools.partial(p_np_score, xs=tuple(xs), y=y, func_index=func_index, clf=clf)

        res = [i for i in pool.map_async(func, nve).get()]
        pool.close()
        pool.join()

        res = np.concatenate(res)

        return res


def p_np_str_name(ve, xns, cns=None, y=None, func_index=None, real_names=None):
    """Batch get name of expression,(without coef and intercept)."""
    _ = y
    func_namesi = [func_names[i] for i in func_index] if func_index is not None else func_names

    if real_names is None:
        xns = [f"x{i}" for i in range(len(xns))]
    else:
        assert len(xns) == real_names
        xns = real_names

    if cns is not None:
        cns = ["{:+.2f}".format(i) for i in cns]
        xns.extend(cns)

    def get_str(vei, n):

        root = vei[0]
        if vei[n] >= 100:
            return xns[vei[n] - 100]
        elif 2 * n >= len(vei):
            return xns[vei[n] - 100]
        else:
            return f"{func_namesi[vei[n]]}({get_str(vei, 2 * n + 1 - root)},{get_str(vei, 2 * n + 2 - root)})"

    res = []
    for vei in ve:
        try:
            resi = get_str(vei, vei[0])
        except TypeError:
            resi = ""
        res.append(resi)

    return res


lr = LinearRegression(fit_intercept=True)
lg = LogisticRegression(fit_intercept=True, penalty='none', dual=False, tol=1e-4, intercept_scaling=1,
                        class_weight=None, random_state=0, solver='lbfgs', max_iter=100,
                        multi_class='auto', verbose=0, warm_start=False, n_jobs=None, l1_ratio=None)



# -*- coding: utf-8 -*-

# @Time     : 2021/11/2 17:23
# @Software : PyCharm
# @License  : GNU General Public License v3.0
# @Author   : xxx
import functools
import pathlib
from multiprocessing import Pool

import numpy as np
from mgetool.draft import DraftPyx

# second
path = pathlib.Path(__file__)
file = path.parent.parent / "source" / "np_tool.pyx"
file = str(file.absolute())
bd = DraftPyx(file, language="c++", temps="np_temps", warm_start=True,log_print=False)
c_numpy_backend_tool = bd.quick_import(build=False)
c_np_score = c_numpy_backend_tool.c_np_score
c_np_cal = c_numpy_backend_tool.c_np_cal


def c_np_score_mp(ve, xs, y, func_index=None, n_jobs=1,clf=False):
    """Batch score with n_jobs."""
    if isinstance(ve, np.ndarray):
        ve = ve.tolist()
    if n_jobs == 1:
        return c_np_score(ve, xs, y, func_index,clf=clf)
    else:

        pool = Pool(n_jobs)

        if func_index is not None:
            func_index = tuple(func_index)

        left = int(len(ve) % n_jobs)

        if left > 0:
            bs = int(len(ve) // n_jobs)
            nve = [ve[bs * (i - 1):i * bs] for i in range(1, n_jobs + 1)]
            nve.append(ve[-left:])
        else:
            bs = int(len(ve) // n_jobs)
            nve = [ve[bs * (i - 1):i * bs] for i in range(1, n_jobs + 1)]

        func = functools.partial(c_np_score, xs=tuple(xs), y=y, func_index=func_index,clf=clf)

        res = [i for i in pool.map_async(func, nve).get()]
        pool.close()
        pool.join()

        res = np.concatenate(res)

        return res



# -*- coding: utf-8 -*-

# @Time     : 2021/11/1 9:19
# @Software : PyCharm
# @License  : GNU General Public License v3.0
# @Author   : xxx

__all__ = ["tools","skflow","sci_formula","gp","cli","backend","source"]


import argparse
import textwrap
from importlib import import_module


class CLIError(Exception):
    """Error for CLI commands.

    A subcommand may raise this.  The message will be forwarded to
    the error() method of the argument args."""


# Important: Following any change to command-line parameters, use
# python3 -m ase.cli.completion to update autocompletion.
commands_ = [

    ('cc_numpy', 'fastgplearn.cli.cc_numpy'),
    ('cc_torch', 'fastgplearn.cli.cc_torch'),

]


def main(prog='fastgplearn', description='fastgplearn command line tool.', args=None):

    commands = commands_
    parser = argparse.ArgumentParser(prog=prog,
                                     description=description,
                                     formatter_class=Formatter)
    parser.add_argument('-T', '--traceback', action='store_true')
    subparsers = parser.add_subparsers(title='Sub-commands',
                                       dest='command')

    subparser = subparsers.add_parser('help',
                                      description='Help',
                                      help='Help for sub-command.')
    subparser.add_argument('helpcommand',
                           nargs='?',
                           metavar='sub-command',
                           help='Provide help for sub-command.')

    functions = {}
    parsers = {}
    for command, module_name in commands:
        cmd = import_module(module_name).CLICommand
        docstring = cmd.__doc__
        if docstring is None:
            short = cmd.short_description
            long = getattr(cmd, 'description', short)
        else:
            parts = docstring.split('\n', 1)
            if len(parts) == 1:
                short = docstring
                long = docstring
            else:
                short, body = parts
                long = short + '\n' + textwrap.dedent(body)
        subparser = subparsers.add_parser(
            command,
            formatter_class=Formatter,
            help=short,
            description=long)
        cmd.add_arguments(subparser)
        functions[command] = cmd.run
        parsers[command] = subparser

    args = parser.parse_args(args)

    if args.command == 'help':
        if args.helpcommand is None:
            parser.print_help()
        else:
            parsers[args.helpcommand].print_help()
    elif args.command is None:
        parser.print_usage()
    else:
        f = functions[args.command]
        try:
            if f.__code__.co_argcount == 1:
                f(args)
            else:
                f(args, parsers[args.command])
        except KeyboardInterrupt:
            pass
        except CLIError as x:
            parser.error(x)
        except Exception as x:
            if args.traceback:
                raise
            else:
                l1 = '{}: {}\n'.format(x.__class__.__name__, x)
                l2 = ('To get a full traceback, use: {} -T {} ...'
                      .format(prog, args.command))
                parser.error(l1 + l2)


class Formatter(argparse.HelpFormatter):
    """Improved help formatter."""

    def _fill_text(self, text, width, indent):
        assert indent == ''
        out = ''
        blocks = text.split('\n\n')
        for block in blocks:
            if block[0] == '*':
                # List items:
                for item in block[2:].split('\n* '):
                    out += textwrap.fill(item,
                                         width=width - 2,
                                         initial_indent='* ',
                                         subsequent_indent='  ') + '\n'
            elif block[0] == ' ':
                # Indented literal block:
                out += block + '\n'
            else:
                # Block of text:
                out += textwrap.fill(block, width=width) + '\n'
            out += '\n'
        return out[:-1]



# -*- coding: utf-8 -*-

# @Time     : 2021/10/8 14:29
# @Software : PyCharm
# @License  : GNU General Public License v3.0
# @Author   : xxx
import os
import pathlib

from mgetool.draft import DraftPyx, TorchJit
from mgetool.imports import BatchFile


class CLICommand:

    """
    Compile pyx or c++ code.

    Example:

        $ fastgplearn cc_torch
    """

    @staticmethod
    def add_arguments(parser):
        pass

    @staticmethod
    def run(args, parser):
        # args = args.parse_args()

        # first
        path = pathlib.Path(__file__)
        file = path.parent.parent / "source" / "torch_tool.cpp"
        file = str(file.absolute())
        bd = TorchJit(file, temps="torch_temp")
        bd.write(functions=["c_torch_score", "c_torch_cal"])
        c_torch_backend_tool = bd.quick_import(build=True, suffix=None)







# -*- coding: utf-8 -*-

# @Time     : 2021/10/31 21:43
# @Software : PyCharm
# @License  : GNU General Public License v3.0
# @Author   : xxx

import pathlib
from multiprocessing import Pool

import numpy as np
import torch
from mgetool.draft import TorchJit


# second
path = pathlib.Path(__file__)
file = path.parent.parent / "source" / "torch_tool.cpp"
file = str(file.absolute())
bd = TorchJit(file,temps="torch_temp", warm_start=True, log_print=False)
c_torch_backend_tool = bd.quick_import(build=False, suffix=None)
c_torch_score0 = c_torch_backend_tool.c_torch_score
c_torch_cal0 = c_torch_backend_tool.c_torch_cal


def c_torch_score(ve, xs, y, func_index, return_numpy=False,clf=False):
    """Batch score."""
    res = c_torch_score0(ve, xs, y, func_index, clf)
    if return_numpy:
        return res.numpy()
    else:
        return res

def c_torch_score_temp(ve, xs, y, func_index,clf=False):
    """Batch score."""
    return c_torch_score0(ve, xs, y, func_index, clf)


def c_torch_cal(ve, xs, y, func_index,clf=False):
    """Batch calculate."""
    res = c_torch_cal0(ve, xs, y, func_index, clf)
    return res


def c_torch_score_mp(ve, xs, y, func_index, n_jobs=1,return_numpy=False,clf=False):
    """Batch score with n_jobs."""
    if isinstance(ve, np.ndarray):
        ve = ve.tolist()
    if n_jobs == 1:
        return c_torch_score(ve, xs, y, func_index, return_numpy,clf)
    else:
        for i in range(3):
            print("For torch with c++ (c_torch), with n_jobs>1, this function is very slow!")

        pool = Pool(n_jobs)

        left = int(len(ve) % n_jobs)

        if left > 0:
            bs = int(len(ve) // n_jobs)
            nve = [ve[bs * (i - 1):i * bs] for i in range(1, n_jobs + 1)]
            nve.append(ve[-left:])
        else:
            bs = int(len(ve) // n_jobs)
            nve = [ve[bs * (i - 1):i * bs] for i in range(1, n_jobs + 1)]

        res = []
        for nvei in nve:
            ret = pool.apply(c_torch_score_temp, (nvei, xs, y, func_index, clf))
            res.append(ret)
        pool.close()
        pool.join()

        if not return_numpy:
            res = torch.cat(res)
        else:
            res = np.concatenate(res)
        return res


# cython: language_level=3
# -*- coding: utf-8 -*-

# @Time     : 2021/11/1 11:08
# @Software : PyCharm
# @License  : GNU General Public License v3.0
# @Author   : xxx

import warnings

import numpy as np

warnings.filterwarnings("ignore")


# def add_(a, b):
#     return np.add(a, b)
#
#
# def sub_(a, b):
#
#     return np.subtract(a, b)
#
#
# def mul_(a, b):
#     return np.multiply(a, b)
#
#
# def div_(a, b):
#     return np.divide(a, b)


def ln_(a, b):
    return np.log(a)


def exp_(a, b):
    return np.exp(a)


def pow2_(a, b):
    return a ** 2


def pow3_(a, b):
    return a ** 3


def rec_(a, b):
    return 1 / a


def max_(a, b):
    return np.max(np.stack((a, b)), 0)


def min_(a, b):
    return np.min(np.stack((a, b)), 0)


def sin_(a, b):
    return np.cos(a)


def cos_(a, b):
    return np.sin(a)


funcs = [np.add, np.subtract, np.multiply, np.divide, ln_, exp_, pow2_, pow3_, rec_, max_, min_, sin_, cos_]
func_names = ["add_", "sub_", "mul_", "div_", "ln_", "exp_", "pow2_", "pow3_", "rec_", "max_", "min_", "sin_", "cos_"]


def get_corr_together(fake_ys, y):
    """


    Args:
        fake_ys (np.ndarray): with shape (n_results, n_sample,).
        y (np.ndarray): with sample (n_sample,).

    Returns:
        corr (np.ndarray): with shape (n_result,)

    """
    # 转置不转置速度相同

    fake_y_mean = np.mean(fake_ys, axis=1)
    y_mean = np.mean(y)

    np.subtract(fake_ys, fake_y_mean.reshape(-1, 1), out=fake_ys)
    y = y - y_mean

    # **2 比 power 块
    corr = np.sum(fake_ys * y, axis=1) / (
            np.sqrt(np.sum(fake_ys ** 2, axis=1)) * np.sqrt(np.sum(y ** 2)))

    return np.abs(np.nan_to_num(corr,posinf=0, neginf=0))


def c_np_cal(ve, xs, y, func_index=None):
    """Batch calculation."""
    funci = [funcs[i] for i in func_index] if func_index is not None else funcs
    error_y = np.zeros_like(y)

    def get_value(vei, n=0):

        root = vei[0]

        if vei[n] >= 100:
            return xs[vei[n] - 100]
        elif 2 * n >= len(vei):
            return xs[vei[n] - 100]
        else:
            return funci[vei[n]](get_value(vei, 2 * n + 1 - root), get_value(vei, 2 * n + 2 - root))

    res = []
    for vei in ve:
        try:
            resi = get_value(vei, vei[0])
        except TypeError:
            resi = error_y
        res.append(resi)

    # res = [get_value(vei, vei[0]) for vei in ve]

    return res


def get_sort_accuracy_together(fake_ys, y):
    """

    Args:
        fake_ys (np.ndarray): with shape (n_results, n_sample,).
        y (np.ndarray): with sample (n_sample,).

    Returns:
        corr (np.ndarray): with shape (n_result,)

    """

    y_sort = np.sort(y)
    y_sort2 = np.sort(y)[::-1]

    fake_ys = np.nan_to_num(fake_ys, nan=np.nan, posinf=np.nan, neginf=np.nan)
    mark = np.any(np.isnan(fake_ys), axis=1)

    fake_ys = np.nan_to_num(fake_ys, nan=-1, posinf=-1, neginf=-1)

    index = np.argsort(fake_ys, axis=1)
    y_pre_sort = y[index]

    acc1 = 1-np.mean(np.abs(y_pre_sort-y_sort), axis=1)
    acc2 = 1-np.mean(np.abs(y_pre_sort-y_sort2), axis=1)

    score = np.max(np.concatenate((acc1.reshape(1,-1),acc2.reshape(1,-1)),axis=0),axis=0)
    score[mark] = 0.0

    return score


def c_np_score(ve, xs, y, func_index,clf=False):
    """Batch score."""
    if isinstance(ve, np.ndarray):
        ve = ve.tolist()
    rs6 = c_np_cal(ve, xs, y, func_index)
    rs7 = np.array(rs6)
    if not clf:
        return get_corr_together(rs7, y)
    else:
        return get_sort_accuracy_together(rs7, y)


cpdef sci_subs(int [:,:] pop,list s,list v,int root,int n,int half_prim_n):
    # """sci subs (pyx version)."""
    cdef float m = 0.5
    cdef int ind
    cdef int site
    for si, vi in zip(s[1:], v[1:]):
        m = -m
        site = int(2 * si + m + 1.5 + root)
        if site >= half_prim_n:
            break
        elif vi != -1:
            ind = int(2 * si + m + 1.5 + root)
            pop[n, ind] = vi





# -*- coding: utf-8 -*-

# @Time     : 2021/10/31 16:13
# @Software : PyCharm
# @License  : GNU General Public License v3.0
# @Author   : xxx


from abc import ABC, abstractmethod

import numpy as np
from mgetool.tool import time_this_function
from sklearn.base import BaseEstimator
from sklearn.metrics import get_scorer

from fastgplearn.backend import backend
from fastgplearn.backend.p_numpy import p_np_str_name, p_np_cal, lg, lr
from fastgplearn.gp import set_seed, generate_random, select_index, mutate_random, crossover, mutate_sci
from fastgplearn.sci_formula import usr_preset
from fastgplearn.tools import Hall, Logs

try:
    import torch
except ImportError:
    torch = None


class SymbolicEstimator(BaseEstimator, ABC):
    def __init__(self, population_size=10000, generations=20, stopping_criteria=0.95, store_of_fame=50,
                 hall_of_fame=3, store=False, p_mutate=0.2, p_crossover=0.5, select_method="tournament",
                 tournament_size=5, device="cpu", sci_preset=None,
                 constant_range=(0, 1.0), constants=None, depth=(2, 5),
                 function_set=('add', 'sub', 'mul', 'div', "pow2", "pow3"),
                 n_jobs=1, verbose=0, random_state=None, method_backend='p_numpy', func_p=None,

                 ):
        """

        Args:
            population_size (int): number of population, default 10000.
            generations (int): number of generations, default 20.
            tournament_size (int): tournament size for selection.
            stopping_criteria (float): criteria of correlation score, max 1.0.
            constant_range (tuple): floats. constant_range=(0,1.0)
            constants (tuple): floats. constants=(-1,1,2,10), if given, The parameter constant_range would be ignored.
            depth (tuple): default (2, 5), The max of depth is not more than 8.
            function_set (tuple): tuple of str. optional: ('add', 'sub', 'mul', 'div', "ln", "exp", "pow2", "pow3", "rec", "max", "min", "sin", "cos").
            n_jobs (int):n jobs to parallel.
            verbose (bool):print message.
            p_mutate: mutate probability.
            p_crossover (float): crossover probability.
            random_state (int):random state
            hall_of_fame (int): hall of frame number to add to next generation.
            store_of_fame (int): hall of frame number to return result.
            method_backend (str): optional: ("p_numpy","c_numpy","p_torch","c_torch")
            device (str): default "cpu", "cuda:0", only accessible of torch.
            func_p (np.ndarray,tuple): with shape (n_function,), probability values of each function.
            sci_preset (str,list): None, "default" or user self-defined list template, default None.


        """

        assert population_size > 100 and generations >= 1
        assert tournament_size <= 30 and hall_of_fame <= store_of_fame
        assert hall_of_fame >= 1
        assert store_of_fame >= 10
        assert depth[1] <= 8
        try:
            self.func_names, self._str_name, self._score_mp, self._score, self._cal = backend[method_backend]
        except KeyError:
            raise NotImplementedError(
                f"The {method_backend} is just accessible after be compiled, see more in 'Install'.")
        self.population_size = self.pop_n = population_size
        self.hall_of_fame = hall_of_fame
        self.generations = self.gen = generations
        self.tournament_size = tournament_size
        self.stopping_criteria = stopping_criteria
        self.constants = constants
        self.depth = depth
        self.select_method = select_method
        self.depth_min = depth[0]
        self.depth_max = depth[1]
        self.function_set = function_set
        self.store_of_fame = store_of_fame

        self.n_jobs = n_jobs
        self.verbose = verbose
        self.random_state = random_state
        self.method_backend = method_backend
        self.p_mutate = p_mutate
        self.p_crossover = p_crossover
        self.constant_n = 5
        self.device = device

        self.constant_gen = []
        self.hall = Hall(size=store_of_fame)
        self.logs = Logs()
        self.store = store

        self.all_scores = []

        if random_state is not None:
            set_seed(random_state)

        self.func_index = [self.func_names.index(f"{i}_") for i in function_set]
        self.func_num = len(function_set)
        self.xs_p = None

        if constants is None:
            self.constant_range = None
        else:
            self.constant_range = constant_range
        self.constants = constants

        if isinstance(func_p, (list, tuple)):
            func_p = list(func_p)
            self.func_p = np.array(func_p)
        else:
            self.func_p = func_p

        self.sci_preset = self.filter_sci_perset(sci_preset)
        self.xs = None
        self.y = None
        self.x_label = None

    def filter_sci_perset(self, sci_preset):
        """Get the available sci available"""
        if sci_preset is None:
            sci_preset = []
        elif sci_preset == "default":
            sci_preset = usr_preset

        temp_sci_preset = []

        func_index_ = set(tuple(self.func_index))
        for site, pre_i in sci_preset:
            if set(tuple(pre_i)) <= func_index_:
                temp_sci_preset.append([list(pre_i), [self.func_index.index(i) for i in pre_i]])

        return temp_sci_preset

    def refresh_xcs_more(self):
        """Refresh X and constant for each generation for torch."""
        xcs, xcs_p, xcs_p_shape0, y = self.refresh_xcs()

        if "torch" in self.method_backend and torch is not None:
            if self.device != "cpu":
                dev = torch.device(self.device)
                xcs = torch.from_numpy(xcs).to(device=dev)
                y = torch.from_numpy(y).to(device=dev)
            else:
                xcs = torch.from_numpy(xcs)
                y = torch.from_numpy(y)

        return xcs, xcs_p, xcs_p_shape0, y

    def refresh_xcs(self):
        """Refresh X and constant for each generation."""
        if self.constants is None and self.constant_range is None:
            temp_cons = np.zeros(self.constant_n)
            self.constant_gen.append(temp_cons)
            return self.xs, None, self.xs.shape[0], self.y

        elif self.constants is None and self.constant_range is not None:
            temp_cons = np.random.random(self.constant_range[0], self.constant_range[1], self.constant_n).astype(
                np.float32)
        else:
            temp_cons = np.array(self.constants).astype(np.float32)
            self.constant_n = temp_cons.shape[0]
        self.constant_gen.append(temp_cons)
        xcs = np.concatenate((self.xs, np.array(temp_cons).reshape(-1, 1)))

        if self.xs_p is not None:
            xcs_p = np.append(0.9 * self.xs_p, 0.1 / np.ones_like(temp_cons))
        else:
            xcs_p = None
        y = self.y

        return xcs, xcs_p, xcs_p.shape[0], y

    @time_this_function
    def fit(self, X: np.ndarray, y: np.ndarray, xs_p: np.ndarray = None, x_label=None):
        """
        Fitting.

        Args:
            X (np.ndarray): with shape (n_sample,n_fea).
            y (np.ndarray): with shape (n_sample,).
            xs_p (np.ndarray): with shape (n_fea,), probability values of each xi.
            x_label (np.ndarray): with shape (n_fea), names of xi.

        """
        self.x_label = x_label
        X, y = self._validate_data(X, y=y)

        if X.ndim == 1:
            raise IndexError("Try to reshape your date to (-1,1), X just accept 2D array.")
        else:
            pass

        self.xs = X.T.astype(np.float32)  # change to (n_sample,n_x)

        self.y = y.ravel().astype(np.float32)

        if isinstance(xs_p, (list, tuple)):
            xs_p = list(xs_p)
            assert len(xs_p) == self.xs.shape[0]
            self.xs_p = np.array(xs_p)
        else:
            self.xs_p = xs_p

        self.hall.refresh_x_num(self.xs.shape[0])
        self.run_gp()

    def single_name(self, n):
        """Get the name of n_ed expression name."""
        n_fea, x_sam = self.xs.shape
        real_names = self.x_label
        if real_names is not None:
            assert len(real_names) == n_fea, "The name of feature are not consist with X."
            xcs = real_names
        else:
            xcs = list(range(n_fea))

        ind, _gen_i, _score, const = self.hall[n]

        if np.all(ind < n_fea):
            cns = None
        else:
            cns = const.tolist()

        return self._single_name(ind.tolist(), xcs, cns=cns, y=self.y, func_index=self.func_index,
                                 real_names=real_names)

    @staticmethod
    def _single_name(vei, xns, cns=None, y=None, func_index=None, real_names=None):
        """Get the name of n_ed expression name."""
        return p_np_str_name([vei, ], xns, cns=cns, y=y, func_index=func_index, real_names=real_names)[0]

    def single_cal(self, n, new_x=None, with_coef=True):
        """Get the temp predict y of n_ed expression name (without coef and intercept),This is not the final result!
        """
        n_fea, x_sam = self.xs.shape
        if new_x is not None:
            assert with_coef is False
            n_fea2, x_sam = new_x.shape
            assert n_fea2 == n_fea, "The numbers of feature for train and test must be same."
            xs = new_x.astype(np.float32)
        else:
            xs = self.xs

        ind, _gen_i, _score, const = self.hall[n]
        if np.all(ind < n_fea):
            xcs = xs
        else:
            xcs = np.concatenate((xs, np.repeat(const.reshape(-1, 1), x_sam, axis=1)), axis=0)

        return self._single_cal(ind.tolist(), xcs, self.y, func_index=self.func_index, with_coef=with_coef)

    @abstractmethod
    def _single_cal(self, vei, xs, y, func_index, with_coef=False):
        """Return tuple(y, coef, intercept). if with coef is False, coef and intercept are all 0."""

    @abstractmethod
    def predict(self, X, y=None, n=0):
        """Return the real predicted y."""

    def score(self, X, y, scoring, n=0):
        """Score."""
        scorer = get_scorer(scoring=scoring)
        func = scorer._score_func
        sign = scorer._sign
        pre_y = self.predict(X, y=None, n=n)
        return sign * func(y, pre_y)

    def run_gp(self):
        """
        Run the GP processing.
        """
        if self.verbose:
            self.logs.print(head=True)

        xcs, xcs_p, xcs_num, y = self.refresh_xcs_more()

        # 1.generate###################################################################

        population = generate_random(self.func_num, xs_num=xcs_num, pop_size=self.pop_n,
                                     depth_min=self.depth_min, depth_max=self.depth_max,
                                     p=None, func_p=self.func_p, xs_p=xcs_p)

        for gen_i in range(1, self.gen + 1):

            # 2.score###################################################################
            population_list = population.tolist()

            score_result = self._score_mp(population_list, xs=xcs, y=y, func_index=self.func_index,
                                          n_jobs=self.n_jobs)

            if "torch" in self.method_backend:
                score_result = score_result.cpu().numpy()

            # 3.log-hall###############################

            if self.store:
                self.all_scores.append(score_result)

            self.hall.update(population, gen_i, score_result, self.constant_gen[-1])
            max_ind_score = np.max(score_result)
            self.logs.record(max_ind_score)
            if self.verbose:
                self.logs.print()
            if max_ind_score >= self.stopping_criteria:
                print("Reach the stopping criteria and terminate early at generation {}".format(gen_i))
                break

            if gen_i == self.gen:
                break

            # 4. next generation ！！！！#######################################################
            # selection and mutate,mate,migration

            if self.constants is None:
                xcs, xcs_p, xcs_num, y = self.refresh_xcs_more()

            couple_index = select_index(score_result, num_percent=self.pop_n - self.hall_of_fame,
                                        method=self.select_method,
                                        tour_num=self.tournament_size)

            couple_index = population[couple_index]

            offspring = crossover(couple_index, p_crossover=self.p_crossover)

            sp = int(0.7 * self.pop_n)

            population[:sp] = mutate_random(offspring[:sp], self.func_num, xcs_num, pop_size=self.pop_n,
                                            depth_min=self.depth_min, depth_max=self.depth_max, p_mutate=self.p_mutate,
                                            p=None, func_p=self.func_p, xs_p=xcs_p)

            population[sp:(self.pop_n - self.hall_of_fame)] = mutate_sci(self.func_num, xcs_num,
                                                                         pop_size=self.pop_n - sp - self.hall_of_fame,
                                                                         depth_min=1, depth_max=5,
                                                                         p=None, func_p=self.func_p, xs_p=xcs_p,
                                                                         sci_preset=self.sci_preset)
            if self.hall_of_fame >= 0:
                population[-self.hall_of_fame:] = self.hall.inds[:self.hall_of_fame]


#
class SymbolicRegressor(SymbolicEstimator):
    """
    A Genetic Programming symbolic regressor.

    A symbolic regressor is an estimator that begins by building a population of naive random formulas to
    represent a relationship. The formulas are represented as tree-like structures with mathematical functions
    being recursively applied to variables and constants. Each successive generation of programs is then evolved
    from the one that came before it by selecting the fittest individuals from the population to undergo genetic
    operations such as crossover, mutation or reproduction.

    The default score for find expression is R (correlation coefficient), Thus this score needs to be further calculated.

    Examples:

    >>> from fastgplearn.skflow import SymbolicRegressor
    >>> est_gp = SymbolicRegressor(population_size=5000,
    ...                     generations=20, stopping_criteria=0.01,
    ...                     p_crossover=0.7, p_mutate_=0.1,
    ...                     max_samples=0.9, verbose=1,
    ...                     random_state=0)
    >>> est_gp.fit(X_train, y_train)
    >>> est_gp.top_n()
    >>> test_score = est_gp.score(X_test,y_test)

    """

    def __init__(self, population_size=10000, generations=20, stopping_criteria=0.95, store_of_fame=50,
                 hall_of_fame=3, store=False, p_mutate=0.2, p_crossover=0.5, select_method="tournament",
                 tournament_size=5, constant_range=(0, 1.0), constants=None, depth=(2, 5),
                 function_set=('add', 'sub', 'mul', 'div', "pow2", "pow3"), sci_preset=None,
                 device="cpu", n_jobs=1, verbose=0, random_state=None, method_backend='p_numpy', func_p=None,
                 ):
        """

        Args:
            population_size (int): number of population, default 10000.
            generations (int): number of generations, default 20.
            tournament_size (int): tournament size for selection.
            stopping_criteria (float): criteria of correlation score, max 1.0.
            constant_range (tuple): floats. constant_range=(0,1.0)
            constants (tuple): floats. constants=(-1,1,2,10), if given, The parameter constant_range would be ignored.
            depth (tuple): default (2, 5), The max of depth is not more than 8.
            function_set (tuple): tuple of str. optional: ('add', 'sub', 'mul', 'div', "ln", "exp", "pow2", "pow3",
                "rec", "max", "min", "sin", "cos").
            n_jobs (int):n jobs to parallel.
            verbose (bool):print message.
            p_mutate: mutate probability.
            p_crossover (float): crossover probability.
            random_state (int):random state
            hall_of_fame (int): hall of frame number to add to next generation.
            store_of_fame (int): hall of frame number to return result.
            method_backend (str): optional: ("p_numpy","c_numpy","p_torch","c_torch")
            device (str): default "cpu", "cuda:0", only accessible of torch.
            func_p (np.ndarray): with shape (n_function,), probability values of each function.
            sci_preset (str,list): None, "default" or user self-defined list template, default None.

        """
        super(SymbolicRegressor, self).__init__(population_size=population_size, generations=generations,
                                                stopping_criteria=stopping_criteria, store=store,
                                                store_of_fame=store_of_fame, random_state=random_state,
                                                hall_of_fame=hall_of_fame, p_mutate=p_mutate, depth=depth,
                                                p_crossover=p_crossover, function_set=function_set,
                                                select_method=select_method, method_backend=method_backend,
                                                tournament_size=tournament_size, sci_preset=sci_preset,
                                                constant_range=constant_range, constants=constants,
                                                device=device, n_jobs=n_jobs, verbose=verbose, func_p=func_p,
                                                )
        self.logs = Logs("Person Corr")

    @staticmethod
    def single_coef_linear(X, y):
        """Fitting by sklearn.linear_model.LinearRegression."""
        lr.fit(X.reshape(-1, 1), y)
        return lr.coef_[0] * X + lr.intercept_, lr.coef_[0], lr.intercept_

    def _single_cal(self, vei, xs, y, func_index, with_coef=False):
        """Return (y,coef,intercept). if with coef is False, coef and intercept are all 0."""

        pre_y = p_np_cal([vei, ], xs, y, func_index)[0]

        if with_coef:
            pre_y_coef, coef_, intercept_ = self.single_coef_linear(pre_y, y)
            return pre_y_coef, coef_, intercept_
        else:
            return pre_y, 0, 0

    def predict(self, X, y=None, n=0):
        """
        Return the real predicted y.

        Args:
            X (np.ndarray): array-like of shape (n_samples, n_features).
            Input vectors, where n_samples is the number of samples and n_features is the number of features
            y (np.ndarray): array-like of shape (n_samples,).
            n (int): calculate by the n_ed expression.

        Returns:
            y (np.ndarray): array-like of shape (n_samples,).
        """
        _ = y
        X = X.T
        _, coef_, intercept_ = self.single_cal(n=n, with_coef=True)
        return self.single_cal(n=n, new_x=X, with_coef=False)[0] * coef_ + intercept_

    def score(self, X, y, scoring="r2", n=0):
        """
        Return the r2 score (default) on the given test data and labels.

        Args:
            X (np.ndarray): array-like of shape (n_samples, n_features).
            y (np.ndarray): array-like of shape (n_samples,).
            scoring (str): see also sklearn.metrics.
            n (int): calculate by the n_ed expression.

        Returns:
            score (float): Mean r2 of ``self.predict(X)`` wrt. `y`.
        """
        return super().score(X, y, scoring, n=n)

    def top_n(self, n=0, scoring="r2"):
        """Print the top n result. The best one is index 0.

        Args:
            scoring (str): see also sklearn.metrics.
            n (int): calculate by the n_ed expression.

        """
        assert self.store_of_fame >= n, "The top_n just accessible while 'hall_of_fame >= n'."
        scorer = get_scorer(scoring=scoring)
        func = scorer._score_func
        sign = scorer._sign

        self.logs.record_and_print(" ")
        self.logs.record_and_print(f"The top {n} result:")
        self.logs.record_and_print(f"Scoring by {scoring}: ( score, expression, coef, intercept )")

        for ni in range(1,n+1):
            pre_y, coef_, intercept_ = self.single_cal(n=n, with_coef=True)
            msg = str((sign * func(self.y, pre_y), self.single_name(ni), coef_, intercept_))
            self.logs.record_and_print(msg)

    def best_expression(self, scoring="r2"):
        """Print the best expression."""
        self.top_n(n=0, scoring=scoring)


class SymbolicClassifier(SymbolicEstimator):
    """
    A Genetic Programming symbolic classifier.

    A symbolic classifier is an estimator that begins by building a population of naive random formulas
    to represent a relationship. The formulas are represented as tree-like structures with mathematical
    functions being recursively applied to variables and constants. Each successive generation of programs
    is then evolved from the one that came before it by selecting the fittest individuals from the population
    to undergo genetic operations such as crossover, mutation or reproduction.

    The default score for find expression is accuracy.

    Examples:

    >>> from fastgplearn.skflow import SymbolicRegressor
    >>> est_gp = SymbolicRegressor(population_size=5000,
    ...                     generations=20, stopping_criteria=0.01,
    ...                     p_crossover=0.7, p_mutate_=0.1,
    ...                     max_samples=0.9, verbose=1,
    ...                     random_state=0)
    >>> est_gp.fit(X_train, y_train)
    >>> est_gp.top_n()
    >>> test_score = est_gp.score(X_test,y_test)
    """

    def __init__(self, population_size=10000, generations=20, stopping_criteria=0.95, store_of_fame=50,
                 hall_of_fame=3, store=False, p_mutate=0.2, p_crossover=0.5, select_method="tournament",
                 tournament_size=5, device="cpu", sci_preset=None,
                 constant_range=(0, 1.0), constants=None, depth=(2, 5),
                 function_set=('add', 'sub', 'mul', 'div', "pow2", "pow3"),
                 n_jobs=1, verbose=0, random_state=None, method_backend='p_numpy', func_p=None,
                 ):
        """

        Args:
            population_size (int): number of population, default 10000.
            generations (int): number of generations, default 20.
            tournament_size (int): tournament size for selection.
            stopping_criteria (float): criteria of correlation score, max 1.0.
            constant_range (tuple): floats. constant_range=(0,1.0)
            constants (tuple): floats. constants=(-1,1,2,10), if given, The parameter constant_range would be ignored.
            depth (tuple): default (2, 5), The max of depth is not more than 8.
            function_set (tuple): tuple of str. optional: ('add', 'sub', 'mul', 'div', "ln", "exp", "pow2", "pow3",
                "rec", "max", "min", "sin", "cos").
            n_jobs (int):n jobs to parallel.
            verbose (bool):print message.
            p_mutate: mutate probability.
            p_crossover (float): crossover probability.
            random_state (int):random state
            hall_of_fame (int): hall of frame number to add to next generation.
            store_of_fame (int): hall of frame number to return result.
            method_backend (str): optional: ("p_numpy","c_numpy","p_torch","c_torch")
            device (str): default "cpu", "cuda:0", only accessible of torch.
            func_p (np.ndarray,tuple): with shape (n_function,), probability values of each function.
            sci_preset (str,list): None, "default" or user self-defined list template, default None.

        """
        super(SymbolicClassifier, self).__init__(population_size=population_size, generations=generations,
                                                 stopping_criteria=stopping_criteria, store=store,
                                                 store_of_fame=store_of_fame, random_state=random_state,
                                                 hall_of_fame=hall_of_fame, p_mutate=p_mutate, depth=depth,
                                                 p_crossover=p_crossover, function_set=function_set,
                                                 select_method=select_method, method_backend=method_backend,
                                                 tournament_size=tournament_size, sci_preset=sci_preset,
                                                 constant_range=constant_range, constants=constants,
                                                 device=device, n_jobs=n_jobs, verbose=verbose, func_p=func_p, )
        self.logs = Logs("Accuracy")

    def fit(self, X: np.ndarray, y: np.ndarray, xs_p: np.ndarray = None, x_label=None):
        """
        Fitting.

        Args:
            X (np.ndarray): with shape (n_sample,n_fea).
            y (np.ndarray): with shape (n_sample,).
            xs_p (np.ndarray): with shape (n_fea,), probability values of each xi.
            x_label (np.ndarray): with shape (n_fea), names of xi.

        """
        class_y = np.unique(y)
        need = {0, 1}
        if class_y.ndim != 2 and set([i for i in class_y]) != need:
            raise ValueError("Just for binary classification problem (with 2 labels: [0,1]).")
        y = y.astype(np.float32)
        super().fit(X, y, xs_p, x_label=x_label)

    def single_coef_logistic(self, X, y):
        """Fitting by sklearn.linear_model.LogisticRegression."""
        lg.fit(X.reshape(-1, 1), y)
        return lg.coef_[0] * X + lg.intercept_, lg.coef_[0], lg.intercept_

    def _single_cal(self, vei, xs, y, func_index, with_coef=False):
        """Return (y,coef,intercept). if with coef is False, coef and intercept are all 0."""

        pre_y = p_np_cal([vei, ], xs, y, func_index)[0]

        if with_coef:
            pre_y_coef, coef_, intercept_ = self.single_coef_logistic(pre_y, self.y)
            return pre_y_coef, coef_, intercept_
        else:
            return pre_y, 0, 0

    @staticmethod
    def cla(pre_y):
        """classification tool."""
        pre_y = 1.0 / (1.0 + np.exp(-pre_y))
        pre_y[np.where(pre_y >= 0.5)] = 1
        pre_y[np.where(pre_y < 0.5)] = 0
        return pre_y

    def predict(self, X, y=None, n=0):
        """
        Return the real predicted y.

        Args:
            X (np.ndarray): array-like of shape (n_samples, n_features).
            Input vectors, where n_samples is the number of samples and n_features is the number of features
            y (np.ndarray): array-like of shape (n_samples,).
            n:

        Returns:
            y (np.ndarray): array-like of shape (n_samples,).
        """

        _ = y

        X = X.T
        _, coef_, intercept_ = self.single_cal(n=n, with_coef=True)
        return self.cla(self.single_cal(n=n, new_x=X, with_coef=False)[0] * coef_ + intercept_)

    def score(self, X, y, scoring="accuracy", n=0):
        """
        Return the mean accuracy on the given test data and labels.

        Args:
            X (np.ndarray): array-like of shape (n_samples, n_features).
            y (np.ndarray): array-like of shape (n_samples,).
            scoring (str): see also sklearn.metrics.
            n (int): calculate by the n_ed expression.

        Returns:
            score (float): Mean accuracy of ``self.predict(X)`` wrt. `y`.
        """
        return super().score(X, y, scoring, n=n)

    def top_n(self, n=0, scoring="accuracy"):
        """Print the top n result. The best one is index 0.

        Args:
            scoring (str): see also sklearn.metrics.
            n (int): calculate by the n_ed expression.

        """

        assert self.store_of_fame >= n, "The top_n just accessible while 'hall_of_fame >= n'."
        scorer = get_scorer(scoring=scoring)
        func = scorer._score_func
        sign = scorer._sign

        self.logs.record_and_print(f"The top {n} result:")
        self.logs.record_and_print(f"Scoring by {scoring}: ( score, expression, coef, intercept )")

        for ni in range(1, n+1):
            pre_y, coef_, intercept_ = self.single_cal(n=n, with_coef=True)
            msg = str((sign * func(self.y, self.cla(pre_y)), self.single_name(ni), coef_, intercept_))
            self.logs.record_and_print(msg)

    def best_expression(self, scoring="accuracy"):
        """Print the best expression."""
        self.top_n(n=0, scoring=scoring)



import unittest
from sklearn.datasets import load_boston
from skflow import SymbolicRegressor

x, y = load_boston(return_X_y=True)



class MyTestCase(unittest.TestCase):
    def setUp(self):
        self.case = object

    def test_something(self):

        sr = SymbolicRegressor(population_size=1000, generations=10, stopping_criteria=0.95,
                               store=False, p_mutate=0.2, p_crossover=0.5, select_method="tournament",
                               tournament_size=5, hall_of_fame=3, store_of_fame=50,
                               constant_range=(0, 1.0), constants=None, depth=(2, 5),
                               function_set=('add', 'sub', 'mul', 'div', "pow2", "pow3", "exp"),
                               n_jobs=1, verbose=True, random_state=0, method_backend='p_numpy', func_p=None,
                               sci_preset="default")
        sr.fit(x, y)
        res = sr.top_n(30)
        res0 = sr.score(x, y, scoring="r2", n=0)
        assert res0>0.5


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

# @Time     : 2021/10/8 14:29
# @Software : PyCharm
# @License  : GNU General Public License v3.0
# @Author   : xxx
import os
import pathlib

from mgetool.draft import DraftPyx
from mgetool.imports import BatchFile


class CLICommand:

    """
    Compile pyx or c++ code.

    Example:

        $ fastgplearn cc_numpy
    """

    @staticmethod
    def add_arguments(parser):
        pass

    @staticmethod
    def run(args, parser):
        # args = args.parse_args()

        # first
        path = pathlib.Path(__file__)
        file = path.parent.parent / "source" / "np_tool.pyx"
        file = str(file.absolute())
        bd = DraftPyx(file, language="c++", temps="np_temps")
        bd.write()
        bd.quick_import(build=True, with_html=True)







# -*- coding: utf-8 -*-

# @Time     : 2021/11/1 9:20
# @Software : PyCharm
# @License  : GNU General Public License v3.0
# @Author   : xxx
import os

import numpy as np
from numpy.random import choice
from numpy import random as nprdm


try:
    from fastgplearn.backend.c_numpy import c_numpy_backend_tool

    csci_subs = c_numpy_backend_tool.sci_subs
except BaseException:
    csci_subs = None


def set_seed(seed):
    """Set random seed."""
    import random
    os.environ['PYTHONHASHSEED'] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    try:
        import torch
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
    except BaseException:
        pass
    try:
        from torch.backends import cudnn
        cudnn.deterministic = True
        cudnn.benchmark = False
    except BaseException:
        pass


def select_index(score, num_percent=0.3, method="tournament", tour_num=3):
    """
    Selection.

    Args:
        score (np.ndarray): score with shape (n_res,)/
        num_percent (int,float): number or percent of population.
        method (str):  "tournament" or "k_best".
        tour_num (int): tournament size .

    Returns:
        index (np.ndarray): index of selection to population.
    """

    size = score.shape[0]
    num_percent = num_percent if num_percent > 1 else int(size * num_percent)

    if method == "k_best":
        return np.argsort(score)[-num_percent:][::-1]
    elif method == "tournament":
        indexx = nprdm.choice(np.arange(size), (tour_num, num_percent))
        scores = score[indexx]
        temp_index = np.argmax(scores, axis=0)[::-1]
        index = indexx[temp_index, np.arange(num_percent)]
        return index
    else:
        raise NotImplementedError


def generate_random(func_num, xs_num, pop_size=10, depth_min=1, depth_max=5, p=None, func_p=None, xs_p=None):
    """
    Generate the first population.
    Each individual with ordered: [mark,1,2,3,4,5,6,7, 103,102,100,102,102,103,102,100]
    1.First part: mark index of root.
    2.second part: index of x gene and f gen.
    3.Third part: protect index of x gene.

    Args:
        func_num (int): func number.
        xs_num (int): x number (n_fea).
        pop_size (int): population size.
        depth_min (int): min depth of expression.
        depth_max (max): max depth of expression.
        p (None): (just for test).
        func_p (np.ndarray): with shape of (n_func), probability,.
        xs_p (np.ndarray): with shape of (n_fea), probability.

    Returns:
        pop (np.ndarray):with shape (n_pop,2**depth_max), population .

    """

    assert depth_max <= 8 and xs_num < 125 and depth_min > 0, "Out of limitation !!!"

    if p is None:
        func_p = np.full(func_num, 0.8 / func_num) if func_p is None else func_p
        xs_p = np.full(xs_num, 0.2 / xs_num) if xs_p is None else xs_p
        p = np.append(func_p, xs_p)

    deps = choice(np.append(np.arange(func_num, dtype=np.uint8), np.arange(100, 100 + xs_num, dtype=np.uint8)),
                  size=(pop_size, 2 ** depth_max), replace=True, p=p)

    deps2 = choice(np.arange(100, 100 + xs_num, dtype=np.uint8), size=(pop_size, 2 ** (depth_max - 1)), replace=True,
                   p=xs_p / np.sum(xs_p))
    deps[:, int(deps.shape[1] / 2):] = deps2

    marks = 1 + 2 ** (depth_max - 1) - 2 ** np.arange(depth_min, depth_max, dtype=np.uint8)
    mark = choice(marks, size=pop_size, replace=True)
    deps[:, 0] = mark

    return deps


def mutate(mutate_pop, func_num, xs_num, depth_min=1, depth_max=5, p_mutate=0.8, p=None, func_p=None, xs_p=None):
    """
    Mutate.
    Each individual with ordered: [mark,1,2,3,4,5,6,7, 103,102,100,102,102,103,102,100]
    1.First part: mark index of root.
    2.second part: index of x gene and f gen.
    3.Third part: protect index of x gene.

    Args:
        func_num (int): func number.
        mutate_pop (np.ndarray): with shape (n_pop,2**depth_max),population.
        xs_num (int): x number (n_fea).
        depth_min (int): min depth of expression.
        depth_max (max): max depth of expression.
        p (None): (just for test).
        func_p (np.ndarray): with shape of (n_func), probability,.
        xs_p (np.ndarray): with shape of (n_fea), probability.
        p_mutate (flaot): probability for mutate.

    Returns:
        pop (np.ndarray): population with shape (n_pop,2**depth_max).

    """
    pop_n, prim_n = mutate_pop.shape
    pii = 1 - (1 - p_mutate) ** (1 / prim_n)
    sub_pop = generate_random(func_num, xs_num, pop_size=pop_n, depth_min=depth_min, depth_max=depth_max,
                              p=p, func_p=func_p, xs_p=xs_p)
    select_mark = choice((0, 1), size=(pop_n, prim_n), replace=True, p=(1 - pii, pii))
    mutate_pop = mutate_pop * (1 - select_mark) + sub_pop * select_mark
    return mutate_pop


def csub_science(pop, sci_preset):
    """
    This would change the init pop!!!

    pyx version for sci substitute.
    """
    now_pop_n, prim_n = pop.shape

    le_scis = len(sci_preset)

    def get_real_index(n, prei=-1):
        s, v = sci_preset[prei]

        root = pop[n, 0]
        pop[n, root] = v[0]
        csci_subs(pop, s, v, root, n)

    index = nprdm.randint(0, high=le_scis, size=now_pop_n)

    for n, i in enumerate(index):
        get_real_index(n=n, prei=i)

    return pop


def sub_science(pop, sci_preset):
    """
    This would change the init pop!!!
    sci substitute.
    """
    now_pop_n, prim_n = pop.shape
    half_prim_n = prim_n / 2

    le_scis = len(sci_preset)

    def get_real_index(n, prei=-1):
        s, v = sci_preset[prei]

        root = pop[n, 0]
        pop[n, root] = v[0]
        m = 0.5
        for si, vi in zip(s[1:], v[1:]):
            m = -m
            site = int(2 * si + m + 1.5 + root)
            if vi != -1 and site < half_prim_n:
                pop[n, 2 * si + int(m + 1.5) + root] = vi

    index = nprdm.randint(0, high=le_scis, size=now_pop_n)

    for n, i in enumerate(index):
        get_real_index(n=n, prei=i)

    return pop


def mutate_random(pop_np, func_num, xs_num, pop_size=10, depth_min=1, depth_max=5, p_mutate=0.8, p=None, func_p=None,
                  xs_p=None):
    """
    Mutate.
    Each individual with ordered: [mark,1,2,3,4,5,6,7, 103,102,100,102,102,103,102,100]
    1.First part: mark index of root.
    2.second part: index of x gene and f gen.
    3.Third part: protect index of x gene.

    Args:
        func_num (int): func number.
        pop_size (int): population size.
        pop_np (np.ndarray): with shape (n_pop,2**depth_max),population.
        xs_num (int): x number (n_fea).
        depth_min (int): min depth of expression.
        depth_max (max): max depth of expression.
        p (None): (just for test).
        func_p (np.ndarray): with shape of (n_func), probability,.
        xs_p (np.ndarray): with shape of (n_fea), probability.
        p_mutate (flaot): probability for mutate.

    Returns:
        pop (np.ndarray): population with shape (n_pop,2**depth_max).

    """
    _ = pop_size

    now_pop_n, prim_n = pop_np.shape
    index = np.arange(now_pop_n)
    nprdm.shuffle(index)
    mutate_pop = pop_np[index]

    mutate_pop = mutate(mutate_pop, func_num, xs_num, depth_min=depth_min, depth_max=depth_max, p_mutate=p_mutate,
                        p=p, func_p=func_p, xs_p=xs_p)

    return mutate_pop


def mutate_sci(func_num, xs_num, pop_size=10, depth_min=1, depth_max=5, p=None, func_p=None,
               xs_p=None, sci_preset=None):
    """
    Mutate.
    Each individual with ordered: [mark,1,2,3,4,5,6,7, 103,102,100,102,102,103,102,100]
    1.First part: mark index of root.
    2.second part: index of x gene and f gen.
    3.Third part: protect index of x gene.

    Args:
        func_num (int): func number.
        pop_size (int): population size.

        xs_num (int): x number (n_fea).
        depth_min (int): min depth of expression.
        depth_max (max): max depth of expression.
        p (None): (just for test).
        func_p (np.ndarray): with shape of (n_func), probability,.
        xs_p (np.ndarray): with shape of (n_fea), probability.
        sci_preset (list of list): the science expression templates.

    Returns:
        pop (np.ndarray): population with shape (n_pop,2**depth_max).

    """
    new_pop = generate_random(func_num, xs_num, pop_size=pop_size, depth_min=depth_min, depth_max=depth_max,
                              p=p, func_p=func_p, xs_p=xs_p)
    if sci_preset is None or sci_preset == []:
        pass
    else:
        if csci_subs is None:
            new_pop = sub_science(new_pop, sci_preset)
        else:
            new_pop = csub_science(new_pop, sci_preset)

    return new_pop


def crossover(pop_np, p_crossover=0.5):
    """
    Corssover.

    Args:
        pop_np (np.ndarray): population
        p_crossover (float): probability for crossover.

    Returns:
        pop (np.ndarray): population with shape (n_pop,2**depth_max).

    """
    now_pop_n, prim_n = pop_np.shape

    whe = int(prim_n / 2)
    need_num = int(now_pop_n * p_crossover / 2)
    need_num2 = need_num * 2

    pop_np[need_num:need_num2, :whe], pop_np[0:need_num, :whe] = pop_np[0:need_num:, :whe], pop_np[need_num:need_num2,
                                                                                            :whe]
    return pop_np



# -*- coding: utf-8 -*-

# @Time     : 2021/11/1 11:07
# @Software : PyCharm
# @License  : GNU General Public License v3.0
# @Author   : xxx

from fastgplearn.backend.p_numpy import p_np_str_name, p_np_cal, p_np_score_mp, p_np_score
from fastgplearn.backend.p_torch import p_torch_score_mp, p_torch_score, p_torch_cal

__all__ = ["c_numpy","c_torch","p_torch","p_numpy"]

func_names = ["add_", "sub_", "mul_", "div_", "ln_", "exp_", "pow2_", "pow3_", "rec_", "max_", "min_", "sin_", "cos_"]

backend = {
    "p_numpy":(func_names, p_np_str_name,p_np_score_mp,p_np_score,p_np_cal),
    "p_torch":(func_names, p_np_str_name,p_torch_score_mp, p_torch_score, p_torch_cal),}
try:
    from fastgplearn.backend.c_numpy import c_np_score_mp, c_np_score, c_np_cal
    backend.update({"c_numpy": (func_names, p_np_str_name, c_np_score_mp, c_np_score, c_np_cal)})
except BaseException:
    pass
try:
    from fastgplearn.backend.c_torch import c_torch_cal, c_torch_score, c_torch_score_mp
    backend.update({"c_torch":(func_names, p_np_str_name,c_torch_score_mp, c_torch_score, c_torch_cal),})
except BaseException:
    pass




